@@ -18,6 +18,7 @@ __PACKAGE__->mk_accessors(
     decrypt
     encrypt
     encryption
+    error
     expiration
     imdsv2
     imdsv2_token
@@ -25,7 +26,9 @@ __PACKAGE__->mk_accessors(
     logger
     order
     passkey
+    print_error
     profile
+    raise_error
     region
     role
     _secret_access_key
@@ -55,37 +58,37 @@ use Scalar::Util qw/reftype/;
 use List::Util qw/pairs any/;
 
 use constant {    ## no critic (ValuesAndExpressions::ProhibitConstantPragma, NamingConventions::Capitalization)
-  AWS_METADATA_BASE_URL            => 'http://169.254.169.254/',
-  AWS_CONTAINER_CREDENTIALS_URL    => 'http://169.254.170.2/',
+  AWS_METADATA_BASE_URL            => q{http://169.254.169.254/},
+  AWS_CONTAINER_CREDENTIALS_URL    => q{http://169.254.170.2/},
   AWS_IAM_SECURITY_CREDENTIALS_URL =>
-    'latest/meta-data/iam/security-credentials/',
-  AWS_AVAILABILITY_ZONE_URL => 'latest/meta-data/placement/availability-zone',
-  COMMA                     => q{,},
-  DEFAULT_CIPHER            => 'Cipher::AES',
-  DEFAULT_SEARCH_ORDER      => 'env,container,role,file',
-  DEFAULT_TIMEOUT           => 3,
-  DEFAULT_WINDOW_INTERVAL   => 5,
-  EMPTY                     => q{},
-  SECONDS_IN_MINUTE         => 60,
-  SECONDS_IN_HOUR           => 3600,
-  INSECURE_MODE             => 2,
-  IMDSv2_URL                => 'latest/api/token',
-  IMDSv2_TTL_HEADER         => 'x-aws-ec2-metadata-token-ttl-seconds',
-  IMDSv2_HEADER             => 'x-aws-ec2-metadata-token',
-  IMDSv2_DEFAULT_TTL        => 21_600,
-  PASSKEY_FORMAT            => '%08X%08x',
-  RANDOM_VALUE              => 0xffffffff,
-  SLASH                     => q{/},
-  TRUE                      => 1,
-  FALSE                     => 0,
-  LOG_FORMAT                => " %s [%s] %s\n",
+    q{latest/meta-data/iam/security-credentials/},
+  AWS_AVAILABILITY_ZONE_URL =>
+    q{latest/meta-data/placement/availability-zone},
+  COMMA                   => q{,},
+  DEFAULT_CIPHER          => q{Cipher::AES},
+  DEFAULT_SEARCH_ORDER    => q{env,container,role,file},
+  DEFAULT_TIMEOUT         => 3,
+  DEFAULT_WINDOW_INTERVAL => 5,
+  EMPTY                   => q{},
+  SECONDS_IN_MINUTE       => 60,
+  SECONDS_IN_HOUR         => 3600,
+  INSECURE_MODE           => 2,
+  IMDSv2_URL              => q{latest/api/token},
+  IMDSv2_TTL_HEADER       => q{x-aws-ec2-metadata-token-ttl-seconds},
+  IMDSv2_HEADER           => q{x-aws-ec2-metadata-token},
+  IMDSv2_DEFAULT_TTL      => 21_600,
+  ISO8601_FORMAT          => q{%Y-%m-%dT%H:%M:%S%z},
+  PASSKEY_FORMAT          => q{%08X%08x},
+  RANDOM_VALUE            => 0xffffffff,
+  SLASH                   => q{/},
+  TRUE                    => 1,
+  FALSE                   => 0,
+  LOG_FORMAT              => qq{ %s [%s] %s\n },
 };
 
-our $VERSION = '1.1.3';
+our $VERSION = '1.1.4';
 
-use vars qw{ @EXPORT_OK };
-
-@EXPORT_OK = qw{ create_passkey };
+our @EXPORT_OK = qw{ create_passkey };
 
 # we only log at debug level, create a default logger
 {
@@ -96,14 +99,18 @@ use vars qw{ @EXPORT_OK };
 
     return if !$self->{'debug'};
 
-    my @tm = localtime time;
-    print {*STDERR} sprintf LOG_FORMAT, strftime( '%c', @tm ),
-      $PROCESS_ID,
-      @message;
+    my @tm        = localtime time;
+    my $timestamp = strftime '%c', @tm;
+
+    my $message = sprintf LOG_FORMAT, $timestamp, $PROCESS_ID, @message;
+
+    print {*STDERR} $message;
   };
 }
 
+########################################################################
 sub new {
+########################################################################
   my ( $class, @args ) = @_;
   my $options = ref $args[0] ? $args[0] : {@args};
 
@@ -146,46 +153,9 @@ sub new {
   return $self;
 } ## end sub new
 
-sub _set_defaults {
-  my ($self) = @_;
-
-  $self->set_debug( $self->get_debug // FALSE );
-
-  $self->set_cache( defined $self->get_cache ? $self->get_cache : TRUE );
-
-  if ( !$self->get_user_agent ) {
-    # set a very low timeout
-    $self->set_user_agent(
-      LWP::UserAgent->new( timeout => $self->get_timeout || DEFAULT_TIMEOUT )
-    );
-  } ## end if ( !$self->get_user_agent)
-
-  my $default_search_order = [ split /\s*,\s*/xsm, DEFAULT_SEARCH_ORDER ];
-
-  if ( !$self->get_order ) {
-    $self->set_order($default_search_order);
-  }
-
-  if ( !ref $self->get_order ) {
-    $self->set_order( [ split /\s*,\s*/xsm, $self->get_order ] );
-  }
-  elsif ( ref $self->get_order && reftype( $self->get_order ) ne 'ARRAY' ) {
-    croak 'order must be a comma delimited string or array ref';
-  }
-
-  foreach my $loc ( @{ $self->get_order } ) {
-    croak "invalid credential location in search order:  [$loc]"
-      if !any {/^$loc$/xsm} @{$default_search_order};
-  }
-
-  if ( !$self->get_profile && any {/file/xsm} @{ $self->get_order } ) {
-    $self->set_profile( $ENV{'AWS_PROFILE'} );
-  }
-
-  return $self;
-} ## end sub _set_defaults
-
+########################################################################
 sub set_default_logger {
+########################################################################
   my ( $self, $debug ) = @_;
 
   $debug = $debug // $self->get_debug;
@@ -197,19 +167,9 @@ sub set_default_logger {
   return $self;
 } ## end sub set_default_logger
 
-sub _init_logger {
-  my ($self) = @_;
-
-  if ( !$self->get_logger || !ref $self->get_logger ) {
-    $self->set_default_logger;
-  }
-
-  $self->get_logger->debug( 'using ' . ref( $self->get_logger ) . ' logger' );
-
-  return $self;
-} ## end sub _init_logger
-
+########################################################################
 sub reset_credentials {
+########################################################################
   my ( $self, $renew ) = @_;
 
   if ( !$renew ) {
@@ -226,7 +186,9 @@ sub reset_credentials {
   return $self;
 } ## end sub reset_credentials
 
+########################################################################
 sub get_default_region {
+########################################################################
   my ($self) = @_;
   my $region;
 
@@ -242,7 +204,7 @@ sub get_default_region {
       $url = _create_metadata_url(AWS_AVAILABILITY_ZONE_URL);
     }
 
-    my $ua = ref($self) ? $self->get_user_agent : LWP::UserAgent->new();
+    my $ua = ref $self ? $self->get_user_agent : LWP::UserAgent->new();
 
     my @headers;
 
@@ -270,33 +232,63 @@ sub get_default_region {
   return $region;
 } ## end sub get_default_region
 
+########################################################################
 sub set_credentials {
+########################################################################
   my ( $self, $creds ) = @_;
 
-  $creds = $creds || $self->find_credentials;
+  $self->set_error(EMPTY);
 
-  if ( $creds->{'aws_secret_access_key'} && $creds->{'aws_access_key_id'} ) {
+  $creds = eval {
+    if ( !$creds ) {
+      $creds = eval { return $self->find_credentials };
+    }
+
+    $self->set_error( $creds->{'error'} || $EVAL_ERROR );
+
+    croak 'no credentials available'
+      if !$creds->{'aws_secret_access_key'} || !$creds->{'aws_access_key_id'};
 
     $self->set_aws_secret_access_key( $creds->{'aws_secret_access_key'} );
+
     $self->set_aws_access_key_id( $creds->{'aws_access_key_id'} );
+
     $self->set_token( $creds->{'token'} );
+
     $self->set_expiration( $creds->{'expiration'} );
-  } ## end if ( $creds->{'aws_secret_access_key'...})
-  else {
-    croak 'no credentials available';
+
+    return $creds;
+  };
+
+  if ( $EVAL_ERROR && !$self->get_error ) {
+    $self->set_error($EVAL_ERROR);
   }
 
+  undef $EVAL_ERROR;
+
+  if ( $self->get_error ) {
+    croak $self->get_error
+      if $self->get_raise_error;
+
+    carp $self->get_error
+      if $self->get_print_error;
+  } ## end if ( $self->get_error )
+
   return $self;
 } ## end sub set_credentials
 
+########################################################################
 sub get_ec2_credentials {
+########################################################################
   goto &find_credentials;
 }
 
+########################################################################
 sub find_credentials {
+########################################################################
   my ( $self, @args ) = @_;
 
-  my $options = ref( $args[0] ) ? $args[0] : {@args};
+  my $options = ref $args[0] ? $args[0] : {@args};
 
   if ( $options->{'profile'} ) {
     $self->set_profile( $options->{'profile'} );
@@ -320,19 +312,9 @@ sub find_credentials {
 
   $self->get_logger->debug( 'search order ' . join COMMA, @search_order );
 
-  my $creds = {};
-
   my %creds_getters = (
     env => sub {
-      if ( $ENV{'AWS_ACCESS_KEY_ID'} && $ENV{'AWS_SECRET_ACCESS_KEY'} ) {
-        @{$creds}{qw/source aws_access_key_id aws_secret_access_key token/}
-          = (
-          'ENV',
-          @ENV{qw/AWS_ACCESS_KEY_ID AWS_SECRET_ACCESS_KEY AWS_SESSION_TOKEN/}
-          );
-      } ## end if ( $ENV{'AWS_ACCESS_KEY_ID'...})
-
-      return $creds;
+      return $self->get_creds_from_env,;
     },
     role => sub {
       return $self->get_creds_from_role;
@@ -345,6 +327,8 @@ sub find_credentials {
     },
   );
 
+  my $creds;
+
   foreach my $location (@search_order) {
 
     $self->get_logger->debug( 'searching for credentials in: ' . $location );
@@ -366,41 +350,44 @@ sub find_credentials {
     }
   } ## end foreach my $k ( keys %{$creds...})
 
+  $self->get_logger->debug( $self->safe_dumper($creds) );
+
   return $creds;
 } ## end sub find_credentials
 
-sub safe_dumper {
-  my ( $self, $obj ) = @_;
+# +------------------+
+# | get_creds_from_* |
+# +------------------+
 
-  return if !ref $obj;
+########################################################################
+sub get_creds_from_env {
+########################################################################
+  my ($self) = @_;
 
-  my $safe_rsp;
+  $self->get_logger->debug('fetching credentials from env');
 
-  if ( $self->get_insecure && $self->get_insecure =~ /^2$/xsm ) {
-    $safe_rsp = Dumper [$obj];
-  }
-  elsif ( $self->get_insecure ) {
-    $safe_rsp = Dumper [$obj];
+  my $creds = {};
 
-    $safe_rsp =~ s/^(.*)(access_?key[^']+)'([^']+)'([^']+)'/$1$2'$3'...'/gxsm;
-    $safe_rsp
-      =~ s/^(.*)(secret_?access[^']+)'([^']+)'([^']+)'/$1$2'$3'...'/gxsm;
-    $safe_rsp =~ s/^(.*)(token[^']+)'([^']+)'([^']+)'/$1$2'$3'...'/gxsm;
-  } ## end elsif ( $self->get_insecure)
-  else {
-    $safe_rsp = '** configuration contents blocked by insecure setting **';
+  my @cred_keys = qw{ source aws_access_key_id aws_secret_access_key token };
+  my @env_keys
+    = qw{ AWS_ACCESS_KEY_ID AWS_SECRET_ACCESS_KEY AWS_SESSION_TOKEN };
+
+  if ( $ENV{'AWS_ACCESS_KEY_ID'} && $ENV{'AWS_SECRET_ACCESS_KEY'} ) {
+    @{$creds}{@cred_keys} = ( 'ENV', @env_keys );
   }
 
-  return $safe_rsp;
-} ## end sub safe_dumper
+  return $creds;
+} ## end sub get_creds_from_env
 
+########################################################################
 sub get_creds_from_process {
+########################################################################
   my ( $self, $process ) = @_;
 
   $self->get_logger->debug("fetching credentials from $process");
 
   open my $fh, q{-|}, $process
-    or die "could not open pipe to $process\n";
+    or croak "could not open pipe to $process\n$OS_ERROR";
 
   local $RS = undef;
 
@@ -408,7 +395,7 @@ sub get_creds_from_process {
   close $fh
     or croak "could not close filehandle on $process\n";
 
-  $credentials = eval { JSON::PP->new->decode($credentials); };
+  $credentials = eval { return decode_json($credentials); };
 
   if ( $EVAL_ERROR || !$credentials ) {
     croak "could not get credentials from process\n$EVAL_ERROR\n";
@@ -433,12 +420,14 @@ sub get_creds_from_process {
     }
   } ## end foreach my $pair ( pairs @credential_keys)
 
-  $credentials->{source} = 'process';
+  $credentials->{'source'} = 'process';
 
   return $credentials;
 } ## end sub get_creds_from_process
 
+########################################################################
 sub get_creds_from_ini_file {
+########################################################################
   my ( $self, $profile ) = @_;
 
   $profile = $profile || $self->get_profile || 'default';
@@ -492,20 +481,23 @@ sub get_creds_from_ini_file {
       $region = $section->{'region'} ? $section->{'region'} : $region;
     }
     elsif ($section) {
+      my @cred_keys = qw{
+        aws_access_key_id
+        aws_secret_access_key
+        token
+        region
+      };
 
-      foreach
-        my $k (qw{aws_access_key_id aws_secret_access_key token region}) {
+      foreach my $k (@cred_keys) {
         if ( defined $section->{$k} ) {
           $creds->{$k} = $section->{$k};
         }
-      } ## end foreach my $k (...)
+      }
 
       $creds->{'source'} = $creds->{'aws_access_key_id'} ? $config : undef;
     } ## end elsif ($section)
   } ## end foreach my $config (qw{ .aws/config .aws/credentials })
 
-  $self->get_logger->debug( $self->safe_dumper($creds) );
-
   $CWD = $cwd;
 
   $self->set_region( $creds->{'region'} || $region );
@@ -513,77 +505,9 @@ sub get_creds_from_ini_file {
   return $creds;
 } ## end sub get_creds_from_ini_file
 
-sub is_token_expired {
-  my ( $self, $window_interval ) = @_;
-  $window_interval = $window_interval // DEFAULT_WINDOW_INTERVAL;
-
-  my $expiration_date = $self->get_expiration();
-
-  my $expired = 0;
-
-  if ( defined $expiration_date ) {
-    # AWS recommends getting credentials 5 minutes prior to expiration
-    my $g = _iso8601_to_time($expiration_date);
-
-    # shave 5 minutes or window interval off of the expiration time
-    $g -= $window_interval * SECONDS_IN_MINUTE;
-
-    # (expiration_time - window_interval) - current_time = # of seconds left before expiration
-    my $seconds_left = $g - time;
-
-    if ( $self->get_debug ) {
-      $self->get_logger->debug("seconds left : $seconds_left");
-      my $hours   = int( $seconds_left / SECONDS_IN_HOUR );
-      my $minutes = int(
-        ( $seconds_left - $hours * SECONDS_IN_HOUR ) / SECONDS_IN_MINUTE );
-      my $seconds = $seconds_left
-        - ( $hours * SECONDS_IN_HOUR + $minutes * SECONDS_IN_MINUTE );
-      $self->get_logger->debug(
-        "$hours hours $minutes minutes $seconds seconds until expiry");
-    } ## end if ( $self->get_debug )
-
-    $expired = ( $seconds_left < 0 ) ? 1 : 0;
-
-    $self->get_logger->debug(
-      Dumper [ 'EXPIRATION TIME: ' . $expiration_date,
-        'EXPIRED: ' . $expired ]
-    );
-  } ## end if ( defined $expiration_date)
-
-  return $expired;
-} ## end sub is_token_expired
-
-sub dump_response {
-  my ( $self, $rsp ) = @_;
-
-  my $safe_rsp;
-
-  if ( $self->get_insecure && $self->get_insecure =~ /^2$/xsm ) {
-    $safe_rsp = $rsp;
-  }
-  elsif ( $self->get_insecure ) {
-    $safe_rsp = {};
-
-    foreach my $k ( keys %{$rsp} ) {
-      if ( $k =~ /content/xsm ) {
-        my $content = $rsp->{$k};
-        $content
-          =~ s/\"(AccessKeyId|Token|SecretAccessKey)\"\s+\:\s+\"[^\"]+\"/\"$1\" : \"...\"/gxsm;
-        $safe_rsp->{$k} = $content;
-      } ## end if ( $k =~ /content/xsm)
-      else {
-        $safe_rsp->{$k} = $rsp->{$k};
-      }
-    } ## end foreach my $k ( keys %{$rsp...})
-  } ## end elsif ( $self->get_insecure)
-  else {
-    $safe_rsp = '** HTTP RESPONSE blocked by insecure setting **';
-  }
-
-  return Dumper [$safe_rsp];
-} ## end sub dump_response
-
+########################################################################
 sub get_creds_from_role {
+########################################################################
   my ($self) = @_;
 
   # try to get credentials from instance role
@@ -592,7 +516,7 @@ sub get_creds_from_role {
   my $ua = $self->get_user_agent;
   my $role;
 
-  my $creds = {};
+  my $creds = { error => undef };
 
   if ( $self->get_imdsv2 ) {
     my $token_url = _create_metadata_url(IMDSv2_URL);
@@ -639,14 +563,17 @@ sub get_creds_from_role {
       # if not 200, then get out of Dodge
       last if !$rsp->is_success;
 
+      my @cred_keys = qw{
+        source role aws_access_key_id aws_secret_access_key
+        token expiration
+      };
+
       if ($role) {
-        my $this = JSON::PP->new->utf8->decode( $rsp->content );
-        @{$creds}{
-          qw/source role aws_access_key_id aws_secret_access_key token expiration/
-          } = (
+        my $this = decode_json( $rsp->content );
+        @{$creds}{@cred_keys} = (
           'IAM', $role,
           @{$this}{qw/AccessKeyId SecretAccessKey Token Expiration/},
-          );
+        );
       } ## end if ($role)
       else {
         $role = $rsp->content;
@@ -662,85 +589,13 @@ sub get_creds_from_role {
   if ($EVAL_ERROR) {
     $creds->{'error'} = $EVAL_ERROR;
   }
-  else {
-    $creds->{'error'} = undef;
-  }
 
   return $creds;
 } ## end sub get_creds_from_role
 
-sub refresh_credentials {
-  goto &refresh_token;
-}
-
-sub refresh_token {
-  my ($self) = @_;
-  my $creds;
-
-  if ( $self->get_container && $self->get_container eq 'ECS' ) {
-    $creds = $self->get_creds_from_container;
-  }
-  elsif ( $self->get_role ) {
-    $creds = $self->get_creds_from_role;
-  }
-
-  croak 'unable to refresh token!'
-    if !ref($creds) || !keys %{$creds};
-
-  return $self->set_credentials($creds);
-} ## end sub refresh_token
-
-sub credential_keys {
-  my ($self) = @_;
-
-  my %credential_keys;
-
-  if ( !$self->get_cache ) {
-    my $creds = $self->find_credentials;
-
-    %credential_keys = (
-      AWS_ACCESS_KEY_ID            => $creds->{'aws_access_key_id'},
-      AWS_SECRET_ACCESS_KEY        => $creds->{'aws_secret_access_key'},
-      AWS_SESSION_TOKEN            => $creds->{'token'},
-      AWS_SESSION_TOKEN_EXPIRATION => $creds->{'expiration'},
-    );
-
-  } ## end if ( !$self->get_cache)
-  else {
-    %credential_keys = (
-      AWS_ACCESS_KEY_ID            => $self->get_aws_access_key_id,
-      AWS_SECRET_ACCESS_KEY        => $self->get_aws_secret_access_key,
-      AWS_SESSION_TOKEN            => $self->get_token,
-      AWS_SESSION_TOKEN_EXPIRATION => $self->get_expiration,
-    );
-  } ## end else [ if ( !$self->get_cache)]
-
-  if ( !defined $credential_keys{AWS_SESSION_TOKEN} ) {
-    delete $credential_keys{AWS_SESSION_TOKEN};
-    delete $credential_keys{AWS_SESSION_TOKEN_EXPIRATION};
-  }
-
-  return \%credential_keys;
-} ## end sub credential_keys
-
-sub as_string {
-  my ($self) = @_;
-
-  return JSON::PP->new->pretty->encode( $self->credential_keys );
-}
-
-sub format_credentials {
-  my ( $self, $format ) = @_;
-
-  $format = $format || "%s %s\n";
-
-  my $credential_keys = $self->credential_keys;
-
-  return join q{}, map { sprintf $format, $_, $credential_keys->{$_} }
-    keys %{$credential_keys};
-} ## end sub format_credentials
-
+########################################################################
 sub get_creds_from_container {
+########################################################################
   my ( $self, $uri ) = @_;
 
   $uri = $uri || $ENV{AWS_CONTAINER_CREDENTIALS_RELATIVE_URI};
@@ -797,7 +652,340 @@ sub get_creds_from_container {
   return $creds;
 } ## end sub get_creds_from_container
 
-sub rotate_credentials {
+# +---------+
+# | DUMPERS |
+# +---------+
+
+########################################################################
+sub safe_dumper {
+########################################################################
+  my ( $self, $obj ) = @_;
+
+  return if !ref $obj;
+
+  my $safe_rsp;
+
+  if ( $self->get_insecure && $self->get_insecure =~ /^2$/xsm ) {
+    $safe_rsp = Dumper [$obj];
+  }
+  elsif ( $self->get_insecure ) {
+    $safe_rsp = Dumper [$obj];
+
+    $safe_rsp =~ s/^(.*)(access_?key[^']+)'([^']+)'([^']+)'/$1$2'$3'...'/gxsm;
+    $safe_rsp
+      =~ s/^(.*)(secret_?access[^']+)'([^']+)'([^']+)'/$1$2'$3'...'/gxsm;
+    $safe_rsp =~ s/^(.*)(token[^']+)'([^']+)'([^']+)'/$1$2'$3'...'/gxsm;
+  } ## end elsif ( $self->get_insecure)
+  else {
+    $safe_rsp = '** configuration contents blocked by insecure setting **';
+  }
+
+  return $safe_rsp;
+} ## end sub safe_dumper
+
+########################################################################
+sub dump_response {
+########################################################################
+  my ( $self, $rsp ) = @_;
+
+  my $safe_rsp;
+
+  if ( $self->get_insecure && $self->get_insecure =~ /^2$/xsm ) {
+    $safe_rsp = $rsp;
+  }
+  elsif ( $self->get_insecure ) {
+    $safe_rsp = {};
+
+    foreach my $k ( keys %{$rsp} ) {
+      if ( $k =~ /content/xsm ) {
+        my $content = $rsp->{$k};
+        $content
+          =~ s/\"(AccessKeyId|Token|SecretAccessKey)\"\s+\:\s+\"[^\"]+\"/\"$1\" : \"...\"/gxsm;
+        $safe_rsp->{$k} = $content;
+      } ## end if ( $k =~ /content/xsm)
+      else {
+        $safe_rsp->{$k} = $rsp->{$k};
+      }
+    } ## end foreach my $k ( keys %{$rsp...})
+  } ## end elsif ( $self->get_insecure)
+  else {
+    $safe_rsp = '** HTTP RESPONSE blocked by insecure setting **';
+  }
+
+  return Dumper [$safe_rsp];
+} ## end sub dump_response
+
+# +---------------+
+# | TOKEN METHODS |
+# +---------------+
+
+########################################################################
+sub is_token_expired {
+########################################################################
+  my ( $self, $window_interval ) = @_;
+  $window_interval = $window_interval // DEFAULT_WINDOW_INTERVAL;
+
+  my $expiration_date = $self->get_expiration();
+
+  my $expired = 0;
+
+  if ( defined $expiration_date ) {
+    # AWS recommends getting credentials 5 minutes prior to expiration
+    my $g = _iso8601_to_time($expiration_date);
+
+    # shave 5 minutes or window interval off of the expiration time
+    $g -= $window_interval * SECONDS_IN_MINUTE;
+
+    # (expiration_time - window_interval) - current_time = # of seconds left before expiration
+    my $seconds_left = $g - time;
+
+    if ( $self->get_debug ) {
+      $self->get_logger->debug("seconds left : $seconds_left");
+      my $hours   = int( $seconds_left / SECONDS_IN_HOUR );
+      my $minutes = int(
+        ( $seconds_left - $hours * SECONDS_IN_HOUR ) / SECONDS_IN_MINUTE );
+      my $seconds = $seconds_left
+        - ( $hours * SECONDS_IN_HOUR + $minutes * SECONDS_IN_MINUTE );
+      $self->get_logger->debug(
+        "$hours hours $minutes minutes $seconds seconds until expiry");
+    } ## end if ( $self->get_debug )
+
+    $expired = ( $seconds_left < 0 ) ? 1 : 0;
+
+    $self->get_logger->debug(
+      Dumper [ 'EXPIRATION TIME: ' . $expiration_date,
+        'EXPIRED: ' . $expired ]
+    );
+  } ## end if ( defined $expiration_date)
+
+  return $expired;
+} ## end sub is_token_expired
+
+########################################################################
+sub refresh_credentials {
+########################################################################
+  goto &refresh_token;
+}
+
+########################################################################
+sub refresh_token {
+########################################################################
+  my ($self) = @_;
+  my $creds;
+
+  if ( $self->get_container && $self->get_container eq 'ECS' ) {
+    $creds = $self->get_creds_from_container;
+  }
+  elsif ( $self->get_role ) {
+    $creds = $self->get_creds_from_role;
+  }
+
+  croak 'unable to refresh token!'
+    if !ref $creds || !keys %{$creds};
+
+  return $self->set_credentials($creds);
+} ## end sub refresh_token
+
+########################################################################
+sub credential_keys {
+########################################################################
+  my ($self) = @_;
+
+  my %credential_keys;
+
+  if ( !$self->get_cache ) {
+    my $creds = $self->find_credentials;
+
+    %credential_keys = (
+      AWS_ACCESS_KEY_ID            => $creds->{'aws_access_key_id'},
+      AWS_SECRET_ACCESS_KEY        => $creds->{'aws_secret_access_key'},
+      AWS_SESSION_TOKEN            => $creds->{'token'},
+      AWS_SESSION_TOKEN_EXPIRATION => $creds->{'expiration'},
+    );
+
+  } ## end if ( !$self->get_cache)
+  else {
+    %credential_keys = (
+      AWS_ACCESS_KEY_ID            => $self->get_aws_access_key_id,
+      AWS_SECRET_ACCESS_KEY        => $self->get_aws_secret_access_key,
+      AWS_SESSION_TOKEN            => $self->get_token,
+      AWS_SESSION_TOKEN_EXPIRATION => $self->get_expiration,
+    );
+  } ## end else [ if ( !$self->get_cache)]
+
+  if ( !defined $credential_keys{AWS_SESSION_TOKEN} ) {
+    delete $credential_keys{AWS_SESSION_TOKEN};
+    delete $credential_keys{AWS_SESSION_TOKEN_EXPIRATION};
+  }
+
+  return \%credential_keys;
+} ## end sub credential_keys
+
+########################################################################
+sub as_string {
+########################################################################
+  my ($self) = @_;
+
+  return JSON::PP->new->pretty->encode( $self->credential_keys );
+} ## end sub as_string
+
+########################################################################
+sub format_credentials {
+########################################################################
+  my ( $self, $format ) = @_;
+
+  $format = $format || "%s %s\n";
+
+  my $credential_keys = $self->credential_keys;
+
+  return join q{}, map { sprintf $format, $_, $credential_keys->{$_} }
+    keys %{$credential_keys};
+} ## end sub format_credentials
+
+# +----------------------------+
+# | CREDENTIALS getters/setter |
+# +----------------------------+
+
+########################################################################
+sub get_aws_access_key_id {
+########################################################################
+  my ($self) = @_;
+
+  if ( !$self->get__access_key_id ) {
+    $self->set_credentials;
+  }
+
+  my $access_key_id
+    = $self->decrypt( $self->get__access_key_id, $self->_fetch_passkey );
+
+  if ( !$self->get_cache ) {
+    $self->set__access_key_id(undef);
+  }
+
+  return $access_key_id;
+} ## end sub get_aws_access_key_id
+
+########################################################################
+sub get_aws_secret_access_key {
+########################################################################
+  my ($self) = @_;
+
+  if ( !$self->get__secret_access_key ) {
+    $self->set_credentials;
+  }
+
+  my $secret_access_key
+    = $self->decrypt( $self->get__secret_access_key, $self->_fetch_passkey );
+
+  if ( !$self->get_cache ) {
+    $self->set__secret_access_key(undef);
+  }
+
+  return $secret_access_key;
+} ## end sub get_aws_secret_access_key
+
+########################################################################
+sub get_token {
+########################################################################
+  my ($self) = @_;
+
+  if ( !$self->get__session_token && $self->get_session_token_required ) {
+    $self->set_credentials;
+  }
+
+  my $token
+    = $self->decrypt( $self->get__session_token, $self->_fetch_passkey );
+
+  if ( !$self->get_cache ) {
+    $self->set__session_token(undef);
+  }
+
+  return $token;
+} ## end sub get_token
+
+########################################################################
+sub set_aws_access_key_id {
+########################################################################
+  my ( $self, $aws_access_key_id, $passkey ) = @_;
+
+  my $key = $aws_access_key_id || undef;
+
+  if ($aws_access_key_id) {
+    $key = $self->encrypt( $aws_access_key_id, $passkey );
+  }
+
+  return $self->set__access_key_id($key);
+} ## end sub set_aws_access_key_id
+
+########################################################################
+sub set_aws_secret_access_key {
+########################################################################
+  my ( $self, $aws_secret_access_key, $passkey ) = @_;
+
+  my $key = $aws_secret_access_key || undef;
+
+  if ($aws_secret_access_key) {
+    $key = $self->encrypt( $aws_secret_access_key, $passkey );
+  }
+
+  return $self->set__secret_access_key($key);
+
+} ## end sub set_aws_secret_access_key
+
+########################################################################
+sub set_token {
+########################################################################
+  my ( $self, $session_token, $passkey ) = @_;
+
+  my $token = $session_token || undef;
+
+  if ($session_token) {
+    $token = $self->encrypt( $session_token, $passkey );
+  }
+
+  return $self->set__session_token($token);
+} ## end sub set_token
+
+# +--------------------+
+# | ENCRYPTION METHODS |
+# +--------------------+
+
+########################################################################
+sub decrypt {
+########################################################################
+  my ( $self, $str, $passkey ) = @_;
+
+  if ( ref $self->get_decrypt && reftype( $self->get_decrypt ) eq 'CODE' ) {
+    return $self->get_decrypt->( $str, $passkey || $self->_fetch_passkey );
+  }
+  else {
+    return $self->_crypt( $str, 'decrypt', $passkey );
+  }
+} ## end sub decrypt
+
+########################################################################
+sub encrypt {
+########################################################################
+  my ( $self, $str, $passkey ) = @_;
+
+  if ( ref $self->get_encrypt && reftype( $self->get_encrypt ) eq 'CODE' ) {
+    return $self->get_encrypt->( $str, $passkey || $self->_fetch_passkey );
+  }
+  else {
+    return $self->_crypt( $str, 'encrypt', $passkey );
+  }
+
+} ## end sub encrypt
+
+########################################################################
+sub create_passkey {
+########################################################################
+  return sprintf PASSKEY_FORMAT, rand RANDOM_VALUE, rand RANDOM_VALUE;
+}
+
+########################################################################
+sub rotate_passkey {
+########################################################################
   my ( $self, $new_passkey ) = @_;
 
   if ( $new_passkey && !ref $new_passkey ) {
@@ -823,101 +1011,79 @@ sub rotate_credentials {
   }
 
   return $new_passkey;
-} ## end sub rotate_credentials
+} ## end sub rotate_passkey
 
-sub get_aws_access_key_id {
-  my ($self) = @_;
-
-  if ( !$self->get__access_key_id ) {
-    $self->set_credentials;
-  }
-
-  my $access_key_id
-    = $self->decrypt( $self->get__access_key_id, $self->_fetch_passkey );
-
-  if ( !$self->get_cache ) {
-    $self->set__access_key_id(undef);
-  }
-
-  return $access_key_id;
-} ## end sub get_aws_access_key_id
-
-sub get_aws_secret_access_key {
-  my ($self) = @_;
-
-  if ( !$self->get__secret_access_key ) {
-    $self->set_credentials;
-  }
-
-  my $secret_access_key
-    = $self->decrypt( $self->get__secret_access_key, $self->_fetch_passkey );
-
-  if ( !$self->get_cache ) {
-    $self->set__secret_access_key(undef);
-  }
-
-  return $secret_access_key;
-} ## end sub get_aws_secret_access_key
-
-sub get_token {
-  my ($self) = @_;
-
-  if ( !$self->get__session_token && $self->get_session_token_required ) {
-    $self->set_credentials;
-  }
-
-  my $token
-    = $self->decrypt( $self->get__session_token, $self->_fetch_passkey );
-
-  if ( !$self->get_cache ) {
-    $self->set__session_token(undef);
-  }
-
-  return $token;
-} ## end sub get_token
-
-sub set_aws_access_key_id {
-  my ( $self, $aws_access_key_id, $passkey ) = @_;
-
-  my $key = $aws_access_key_id || undef;
-
-  if ($aws_access_key_id) {
-    $key = $self->encrypt( $aws_access_key_id, $passkey );
-  }
-
-  return $self->set__access_key_id($key);
-} ## end sub set_aws_access_key_id
-
-sub set_aws_secret_access_key {
-  my ( $self, $aws_secret_access_key, $passkey ) = @_;
-
-  my $key = $aws_secret_access_key || undef;
-
-  if ($aws_secret_access_key) {
-    $key = $self->encrypt( $aws_secret_access_key, $passkey );
-  }
-
-  return $self->set__secret_access_key($key);
-
-} ## end sub set_aws_secret_access_key
-
-sub set_token {
-  my ( $self, $session_token, $passkey ) = @_;
-
-  my $token = $session_token || undef;
-
-  if ($session_token) {
-    $token = $self->encrypt( $session_token, $passkey );
-  }
-
-  return $self->set__session_token($token);
-} ## end sub set_token
+########################################################################
+sub rotate_credentials {
+########################################################################
+  goto &rotate_passkey;
+}
 
 # +-----------------+
 # | PRIVATE METHODS |
 # +-----------------+
 
+########################################################################
+sub _init_logger {
+########################################################################
+  my ($self) = @_;
+
+  if ( !$self->get_logger || !ref $self->get_logger ) {
+    $self->set_default_logger;
+  }
+
+  $self->get_logger->debug( 'using ' . ref( $self->get_logger ) . ' logger' );
+
+  return $self;
+} ## end sub _init_logger
+
+########################################################################
+sub _set_defaults {
+########################################################################
+  my ($self) = @_;
+
+  $self->set_debug( $self->get_debug // FALSE );
+
+  $self->set_cache( defined $self->get_cache ? $self->get_cache : TRUE );
+
+  if ( !$self->get_user_agent ) {
+    # set a very low timeout
+    $self->set_user_agent(
+      LWP::UserAgent->new( timeout => $self->get_timeout || DEFAULT_TIMEOUT )
+    );
+  } ## end if ( !$self->get_user_agent)
+
+  my $default_search_order = [ split /\s*,\s*/xsm, DEFAULT_SEARCH_ORDER ];
+
+  if ( !$self->get_order ) {
+    $self->set_order($default_search_order);
+  }
+
+  if ( !ref $self->get_order ) {
+    $self->set_order( [ split /\s*,\s*/xsm, $self->get_order ] );
+  }
+  elsif ( ref $self->get_order && reftype( $self->get_order ) ne 'ARRAY' ) {
+    croak 'order must be a comma delimited string or array ref';
+  }
+
+  foreach my $loc ( @{ $self->get_order } ) {
+    croak "invalid credential location in search order:  [$loc]"
+      if !any {/^$loc$/xsm} @{$default_search_order};
+  }
+
+  if ( !$self->get_profile && any {/file/xsm} @{ $self->get_order } ) {
+    $self->set_profile( $ENV{'AWS_PROFILE'} );
+  }
+
+  $self->set_raise_error( $self->get_raise_error // TRUE );
+  $self->set_print_error( $self->get_print_error // TRUE );
+
+  return $self;
+} ## end sub _set_defaults
+
+########################################################################
 sub _fetch_passkey {
+########################################################################
   my ($self) = @_;
 
   my $passkey = eval {
@@ -932,7 +1098,9 @@ sub _fetch_passkey {
   return $passkey;
 } ## end sub _fetch_passkey
 
+########################################################################
 sub _init_encryption {
+########################################################################
   my ($self) = @_;
 
   # if one is set, both must be set
@@ -970,7 +1138,9 @@ sub _init_encryption {
   return $self->get_encryption;
 } ## end sub _init_encryption
 
+########################################################################
 sub _crypt {
+########################################################################
   my ( $self, $str, $op, $passkey ) = @_;
 
   return if !$str;
@@ -984,73 +1154,46 @@ sub _crypt {
       '-pass'        => $passkey,
       '-key'         => $passkey,
       '-cipher'      => $self->get_cipher,
-      '-nodeprecate' => 1,
+      '-nodeprecate' => TRUE,
     );
   } ## end if ( $self->get_encryption)
 
   # at least obfuscate the credentials
   if ( $op eq 'decrypt' ) {
-    $str
-      = ref($cipher)
-      ? $cipher->decrypt( decode_base64($str) )
-      : decode_base64($str);
-  } ## end if ( $op eq 'decrypt' )
+    $str = decode_base64($str);
+    $str = ref $cipher ? $cipher->decrypt($str) : $str;
+  }
   else {
-    $str
-      = ref($cipher)
-      ? encode_base64( $cipher->encrypt($str) )
-      : encode_base64($str);
-  } ## end else [ if ( $op eq 'decrypt' )]
+    $str = ref $cipher ? $cipher->encrypt($str) : $str;
+    $str = encode_base64($str);
+  }
 
   return $str;
 } ## end sub _crypt
 
-sub decrypt {
-  my ( $self, $str, $passkey ) = @_;
-
-  if ( ref $self->get_decrypt && reftype( $self->get_decrypt ) eq 'CODE' ) {
-    return $self->get_decrypt->( $str, $passkey || $self->_fetch_passkey );
-  }
-  else {
-    return $self->_crypt( $str, 'decrypt', $passkey );
-  }
-} ## end sub decrypt
-
-sub encrypt {
-  my ( $self, $str, $passkey ) = @_;
-
-  if ( ref $self->get_encrypt && reftype( $self->get_encrypt ) eq 'CODE' ) {
-    return $self->get_encrypt->( $str, $passkey || $self->_fetch_passkey );
-  }
-  else {
-    return $self->_crypt( $str, 'encrypt', $passkey );
-  }
-
-} ## end sub encrypt
-
-sub create_passkey {
-  return sprintf PASSKEY_FORMAT, rand RANDOM_VALUE, rand RANDOM_VALUE;
-}
-
+########################################################################
 sub _iso8601_to_time {
+########################################################################
   my $iso8601 = shift;
 
   $iso8601 =~ s/^(.*)Z$/$1\+00:00/xsm;
 
   my $gmtime = eval {
-    local $ENV{TZ} = 'GMT';
+    local $ENV{'TZ'} = 'GMT';
 
-    timegm( strptime( $iso8601, '%Y-%m-%dT%H:%M:%S%z' ) );
+    timegm( strptime( $iso8601, ISO8601_FORMAT ) );
   };
 
   return $gmtime;
 } ## end sub _iso8601_to_time
 
+########################################################################
 sub _create_metadata_url {
+########################################################################
   my ($url) = @_;
 
   return AWS_METADATA_BASE_URL . $url;
-}
+} ## end sub _create_metadata_url
 
 1;
 
@@ -1234,6 +1377,13 @@ new passkey method.
  $credentials->reset_credentials(1);
 
 
+=item print_error
+
+Whether to print the error if no credenials are found. C<raise_error>
+implies C<print_error>.
+
+default: true
+
 =item profile
 
 The profile name in the configuration file (F<~/.aws/config> or
@@ -1246,6 +1396,12 @@ so you can invoke your script like this:
 
  $ AWS_PROFILE=sandbox my-script.pl
 
+=item raise_error
+
+Whether to raise an error if credentials are not found.
+
+default: true
+
 =item region
 
 Default region. The class will attempt to find the region in either
